@@ -1,3 +1,5 @@
+// SPDX-License-Identifier: Apache-2.0
+// 
 // Copyright 2008-2016 Conrad Sanderson (http://conradsanderson.id.au)
 // Copyright 2008-2016 National ICT Australia (NICTA)
 // 
@@ -24,7 +26,7 @@ inline
 void
 op_real::apply( Mat<typename T1::pod_type>& out, const mtOp<typename T1::pod_type, T1, op_real>& X )
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
   typedef typename T1::pod_type T;
   
@@ -67,7 +69,7 @@ inline
 void
 op_real::apply( Cube<typename T1::pod_type>& out, const mtOpCube<typename T1::pod_type, T1, op_real>& X )
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
   typedef typename T1::pod_type T;
   
@@ -112,12 +114,17 @@ inline
 void
 op_imag::apply( Mat<typename T1::pod_type>& out, const mtOp<typename T1::pod_type, T1, op_imag>& X )
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
-  typedef typename T1::pod_type T;
+  typedef typename T1::elem_type eT;
+  typedef typename T1::pod_type   T;
   
   const Proxy<T1> P(X.m);
   
+  if(is_cx<eT>::no)  { out.zeros(P.get_n_rows(), P.get_n_cols()); return; }
+  
+  // aliasing not possible at this point, as eT must be std::complex
+  
   const uword n_rows = P.get_n_rows();
   const uword n_cols = P.get_n_cols();
     
@@ -134,7 +141,7 @@ op_imag::apply( Mat<typename T1::pod_type>& out, const mtOp<typename T1::pod_typ
     
     for(uword i=0; i < n_elem; ++i)
       {
-      out_mem[i] = std::imag( A[i] );
+      out_mem[i] = access::tmp_imag( A[i] );
       }
     }
   else
@@ -142,7 +149,7 @@ op_imag::apply( Mat<typename T1::pod_type>& out, const mtOp<typename T1::pod_typ
     for(uword col=0; col < n_cols; ++col)
     for(uword row=0; row < n_rows; ++row)
       {
-      *out_mem = std::imag( P.at(row,col) );
+      *out_mem = access::tmp_imag( P.at(row,col) );
       out_mem++;
       }
     }
@@ -155,12 +162,17 @@ inline
 void
 op_imag::apply( Cube<typename T1::pod_type>& out, const mtOpCube<typename T1::pod_type, T1, op_imag>& X )
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
-  typedef typename T1::pod_type T;
+  typedef typename T1::elem_type eT;
+  typedef typename T1::pod_type   T;
   
   const ProxyCube<T1> P(X.m);
   
+  if(is_cx<eT>::no)  { out.zeros(P.get_n_rows(), P.get_n_cols(), P.get_n_slices()); return; }
+  
+  // aliasing not possible at this point, as eT must be std::complex
+  
   const uword n_rows   = P.get_n_rows();
   const uword n_cols   = P.get_n_cols();
   const uword n_slices = P.get_n_slices();
@@ -178,7 +190,7 @@ op_imag::apply( Cube<typename T1::pod_type>& out, const mtOpCube<typename T1::po
     
     for(uword i=0; i < n_elem; ++i)
       {
-      out_mem[i] = std::imag( A[i] );
+      out_mem[i] = access::tmp_imag( A[i] );
       }
     }
   else
@@ -187,7 +199,7 @@ op_imag::apply( Cube<typename T1::pod_type>& out, const mtOpCube<typename T1::po
     for(uword col=0;   col   < n_cols;   ++col  )
     for(uword row=0;   row   < n_rows;   ++row  )
       {
-      *out_mem = std::imag( P.at(row,col,slice) );
+      *out_mem = access::tmp_imag( P.at(row,col,slice) );
       out_mem++;
       }
     }
@@ -200,7 +212,7 @@ inline
 void
 op_abs::apply( Mat<typename T1::pod_type>& out, const mtOp<typename T1::pod_type, T1, op_abs>& X )
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
   typedef typename T1::pod_type T;
   
@@ -208,7 +220,7 @@ op_abs::apply( Mat<typename T1::pod_type>& out, const mtOp<typename T1::pod_type
   
   const uword n_rows = P.get_n_rows();
   const uword n_cols = P.get_n_cols();
-    
+  
   out.set_size(n_rows, n_cols);
   
   T* out_mem = out.memptr();
@@ -220,10 +232,23 @@ op_abs::apply( Mat<typename T1::pod_type>& out, const mtOp<typename T1::pod_type
     const uword   n_elem  = P.get_n_elem();
           ea_type A       = P.get_ea();
     
-    for(uword i=0; i < n_elem; ++i)
+    #if defined(ARMA_USE_OPENMP)
       {
-      out_mem[i] = std::abs( A[i] );
+      const int n_threads = mp_thread_limit::get();
+      #pragma omp parallel for schedule(static) num_threads(n_threads)
+      for(uword i=0; i < n_elem; ++i)
+        {
+        out_mem[i] = std::abs( A[i] );
+        }
       }
+    #else
+      {
+      for(uword i=0; i < n_elem; ++i)
+        {
+        out_mem[i] = std::abs( A[i] );
+        }
+      }
+    #endif
     }
   else
     {
@@ -243,7 +268,7 @@ inline
 void
 op_abs::apply( Cube<typename T1::pod_type>& out, const mtOpCube<typename T1::pod_type, T1, op_abs>& X )
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
   typedef typename T1::pod_type T;
   
@@ -264,10 +289,23 @@ op_abs::apply( Cube<typename T1::pod_type>& out, const mtOpCube<typename T1::pod
     const uword   n_elem  = P.get_n_elem();
           ea_type A       = P.get_ea();
     
-    for(uword i=0; i < n_elem; ++i)
+    #if defined(ARMA_USE_OPENMP)
       {
-      out_mem[i] = std::abs( A[i] );
+      const int n_threads = mp_thread_limit::get();
+      #pragma omp parallel for schedule(static) num_threads(n_threads)
+      for(uword i=0; i < n_elem; ++i)
+        {
+        out_mem[i] = std::abs( A[i] );
+        }
       }
+    #else
+      {
+      for(uword i=0; i < n_elem; ++i)
+        {
+        out_mem[i] = std::abs( A[i] );
+        }
+      }
+    #endif
     }
   else
     {
@@ -288,7 +326,7 @@ inline
 void
 op_arg::apply( Mat<typename T1::pod_type>& out, const mtOp<typename T1::pod_type, T1, op_arg>& X )
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
   typedef typename T1::elem_type eT;
   typedef typename T1::pod_type   T;
@@ -332,7 +370,7 @@ inline
 void
 op_arg::apply( Cube<typename T1::pod_type>& out, const mtOpCube<typename T1::pod_type, T1, op_arg>& X )
   {
-  arma_extra_debug_sigprint();
+  arma_debug_sigprint();
   
   typedef typename T1::elem_type eT;
   typedef typename T1::pod_type   T;
